"""Provides:
  1. Toolkit for testing factor strategies
  2. A lot of financial metrics to assess portfolios performance
  3. Fancy visualization of results

Source: https://github.com/atomtosov/pqr/

Affiliation: https://fmlab.hse.ru/
"""

from .core.universe import *
from .core.benchmark import *
from .core.factor import *
from .core.portfolio import *
from .core import factor_model as fm

from .analytics import dashboards as dash
from .analytics import metrics
from .analytics import tests
from .analytics import regressions

from . import utils



from __future__ import annotations

from typing import Sequence, Protocol, Optional

import matplotlib.pyplot as plt
import pandas as pd
from IPython.display import display

from pqr.core import Portfolio, Benchmark

__all__ = [
    "Dashboard",
    "Graph",
    "Table",
]


class NumericMetric(Protocol):
    def fancy(self, portfolio: Portfolio) -> str:
        pass

    @property
    def fancy_name(self, portfolio: Portfolio) -> str:
        pass


class TimeSeriesMetric(Protocol):
    def __call__(self, portfolio: Portfolio) -> pd.Series:
        pass

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        pass

    @property
    def fancy_name(self, portfolio: Portfolio) -> str:
        pass


class Dashboard:
    def __init__(self, *items: Graph | Table):
        self.items = items

    def __call__(self, portfolios: Sequence[Portfolio]) -> None:
        for item in self.items:
            item(portfolios)


class Graph:
    def __init__(
            self,
            metric: TimeSeriesMetric,
            benchmark: Optional[Benchmark] = None,
            log_scale: bool = False,
            figsize: tuple[int, int] = (10, 10)
    ):
        self.metric = metric
        self.benchmark = benchmark
        self.log_scale = log_scale
        self.figsize = figsize

    def __call__(self, portfolios: Sequence[Portfolio]) -> None:
        plt.figure(figsize=self.figsize)

        for portfolio in portfolios:
            plt.plot(
                self.metric(portfolio) if self.log_scale else
                self.metric.fancy(portfolio),
                label=portfolio.name
            )

        if self.benchmark is not None:
            starts_from = min(portfolio.returns.index[0] for portfolio in portfolios)
            benchmark_returns = self.benchmark.returns[starts_from:]
            benchmark_returns.iloc[0] = 0.0
            benchmark = Benchmark(
                benchmark_returns,
                name=self.benchmark.name
            )
            plt.plot(
                self.metric(benchmark) if self.log_scale else
                self.metric.fancy(benchmark),
                label=self.benchmark.name,
                color="gray",
                alpha=0.8,
            )

        if self.log_scale:
            plt.yscale("symlog")

        plt.title(f"Portfolios {self.metric.fancy_name}")
        plt.xlabel("Date")
        plt.ylabel(f"{self.metric.fancy_name}{' (log scale)' if self.log_scale else ''}")
        plt.legend()
        plt.grid()

        plt.show()


class Table:
    def __init__(self, *metrics: NumericMetric):
        self.metrics = metrics

    def __call__(self, portfolios: Sequence[Portfolio]) -> None:
        metrics_table = {}
        for metric in self.metrics:
            metrics_table[metric.fancy_name] = [
                metric.fancy(portfolio) for portfolio in portfolios
            ]

        display(
            pd.DataFrame(
                metrics_table,
                index=[portfolio.name for portfolio in portfolios]
            ).T
        )



import numpy as np
import pandas as pd

from pqr.utils import align_many


class Universe:
    def __init__(self, prices: pd.DataFrame):
        self.prices = prices.astype(float)
        self.mask = prices.notnull()

    def filter(self, mask: pd.DataFrame):
        self.mask &= mask

    def __call__(self, positions: pd.DataFrame) -> pd.Series:
        prices, mask, positions = align_many(self.prices, self.mask, positions)

        universe_returns = prices.pct_change().to_numpy()[1:]

        positions_available = positions.to_numpy()[:-1]
        portfolio_returns = np.where(
            mask.to_numpy()[:-1],
            (positions_available * universe_returns), 0
        )

        dead_returns = np.where(
            np.isnan(portfolio_returns) & ~np.isclose(positions_available, 0),
            -positions_available, 0
        )
        returns = np.nansum(portfolio_returns, axis=1) + np.nansum(dead_returns, axis=1)

        return pd.Series(
            np.insert(returns, 0, 0),
            index=positions.index.copy()
        )



from __future__ import annotations

import functools as ft
from typing import Callable, Literal, Sequence, Optional

import numpy as np
import numpy.typing as npt
import pandas as pd

from pqr.utils import align
from .utils import compose, top_single, bottom_single

__all__ = [
    "Factor",
    "Preprocessor",

    "Filter",
    "LookBackPctChange",
    "LookBackMean",
    "LookBackMedian",
    "LookBackMin",
    "LookBackMax",
    "Lag",
    "Hold",
    "ReplaceWithNan",
]

Preprocessor = Callable[[pd.DataFrame], pd.DataFrame]


class Factor:
    def __init__(
            self,
            values: pd.DataFrame,
            better: Literal["more", "less"],
            preprocessor: Optional[Preprocessor | Sequence[Preprocessor]] = None,
    ):
        self.values = values.astype(float)
        if isinstance(preprocessor, Sequence):
            self.values = compose(*preprocessor)(self.values)
        elif preprocessor is not None:
            self.values = preprocessor(self.values)

        self.better = better

    def is_better_more(self) -> bool:
        return self.better == "more"

    def is_better_less(self) -> bool:
        return self.better == "less"

    def quantile(self, q: npt.ArrayLike[float]) -> pd.DataFrame:
        q = np.array(q).reshape((-1,))

        quantiles = np.nanquantile(
            self.values.to_numpy(),
            q if self.is_better_less() else 1 - q,
            axis=1
        ).T

        return pd.DataFrame(
            quantiles,
            index=self.values.index,
            columns=[f"q_{_q:.2f}" for _q in q]
        )

    def top(self, n: npt.ArrayLike[int]) -> pd.DataFrame:
        n = np.array(n).reshape((-1,))

        top_func = ft.partial(
            top_single if self.is_better_more() else bottom_single,
            n=n
        )

        return pd.DataFrame(
            np.apply_along_axis(top_func, 1, self.values.to_numpy()),
            index=self.values.index,
            columns=[f"top_{_n}" for _n in n]
        )

    def bottom(self, n: npt.ArrayLike[int]) -> pd.DataFrame:
        n = np.array(n).reshape((-1,))

        bottom_func = ft.partial(
            bottom_single if self.is_better_more() else top_single,
            n=n
        )

        return pd.DataFrame(
            np.apply_along_axis(bottom_func, 1, self.values.to_numpy()),
            index=self.values.index,
            columns=[f"bottom_{_n}" for _n in n]
        )


class Filter:
    def __init__(self, mask: pd.DataFrame):
        self.mask = mask.astype(bool)

    def __call__(self, values: pd.DataFrame) -> pd.DataFrame:
        universe, values = align(self.mask, values)
        return pd.DataFrame(
            np.where(universe.to_numpy(), values.to_numpy(), np.nan),
            index=values.index,
            columns=values.columns
        )


class LookBackPctChange:
    def __init__(self, period: int):
        self.period = period

    def __call__(self, values: pd.DataFrame) -> pd.DataFrame:
        return values.pct_change(self.period).iloc[self.period:]


class LookBackMean:
    def __init__(self, period: int):
        self.period = period

    def __call__(self, values: pd.DataFrame) -> pd.DataFrame:
        return values.rolling(self.period + 1, axis=0).mean().iloc[self.period:]


class LookBackMedian:
    def __init__(self, period: int):
        self.period = period

    def __call__(self, values: pd.DataFrame) -> pd.DataFrame:
        return values.rolling(self.period + 1, axis=0).median().iloc[self.period:]


class LookBackMin:
    def __init__(self, period: int):
        self.period = period

    def __call__(self, values: pd.DataFrame) -> pd.DataFrame:
        return values.rolling(self.period + 1, axis=0).min().iloc[self.period:]


class LookBackMax:
    def __init__(self, period: int):
        self.period = period

    def __call__(self, values: pd.DataFrame) -> pd.DataFrame:
        return values.rolling(self.period + 1, axis=0).max().iloc[self.period:]


class Lag:
    def __init__(self, period: int):
        self.period = period

    def __call__(self, values: pd.DataFrame) -> pd.DataFrame:
        if self.period == 0:
            return values

        return pd.DataFrame(
            values.to_numpy()[:-self.period],
            index=values.index[self.period:].copy(),
            columns=values.columns.copy()
        )


class Hold:
    def __init__(self, period: int):
        self.period = period

    def __call__(self, values: pd.DataFrame) -> pd.DataFrame:
        periods = np.zeros(len(values), dtype=int)
        update_periods = np.arange(len(values), step=self.period)
        periods[update_periods] = update_periods
        update_mask = np.maximum.accumulate(periods[:, np.newaxis], axis=0)

        return pd.DataFrame(
            np.take_along_axis(values.to_numpy(), update_mask, axis=0),
            index=values.index.copy(),
            columns=values.columns.copy()
        )


class ReplaceWithNan:
    def __init__(self, to_replace):
        self.to_replace = to_replace

    def __call__(self, values: pd.DataFrame) -> pd.DataFrame:
        return values.replace(self.to_replace, np.nan)



from __future__ import annotations

from typing import Callable, Optional, Literal, Sequence

import numpy as np
import pandas as pd
from scipy.stats import ttest_1samp

import pqr
from pqr.core import Portfolio, AllocationStep, Universe, Hold
from pqr.utils import align_many
from .utils import estimate_holding

__all__ = [
    "TTest",
    "ZeroIntelligenceTest",
    "OpportunitiesTest"
]


class TTest:
    def __init__(
            self,
            h0: float = 0.0,
            alternative: str = "greater"
    ):
        self.h0 = h0
        self.alternative = alternative

    def __call__(self, portfolio: Portfolio):
        return ttest_1samp(portfolio.returns, self.h0, alternative=self.alternative)


class ZeroIntelligenceTest:
    def __init__(
            self,
            universe: Universe,
            allocation_strategy: AllocationStep | Sequence[AllocationStep],
            target_metric: Callable[[Portfolio], pd.Series],
            holding: Optional[int] = None,
            n: int = 100,
            seed: Optional[int] = None,
    ):
        self.universe = universe
        self.allocation_strategy = allocation_strategy
        self.target_metric = target_metric
        self.n = n
        self.holding = holding
        self.seed = seed

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        random_picking = RandomPicking(
            self.universe,
            portfolio,
            self.n,
            self.holding,
            np.random.default_rng(self.seed)
        )

        random_metrics = []
        for longs, shorts in random_picking:
            random_portfolio = Portfolio(
                self.universe,
                longs=longs,
                shorts=shorts,
                allocation_strategy=self.allocation_strategy,
            )

            random_metrics.append(
                self.target_metric(random_portfolio).to_numpy()
            )

        random_metrics = np.array(random_metrics).T
        portfolio_metric = self.target_metric(portfolio)

        return pd.Series(
            np.nanmean(
                random_metrics <= portfolio_metric.to_numpy()[:, np.newaxis],
                axis=1
            ),
            index=portfolio_metric.index
        )


class RandomPicking:
    def __init__(
            self,
            universe: Universe,
            portfolio: Portfolio,
            n: int = 100,
            holding: Optional[int] = None,
            rng: Optional[np.random.Generator] = None,
    ):
        self.portfolio = portfolio
        self.n = n
        self.holding = holding if holding is not None else estimate_holding(portfolio.picks)
        self.rng = rng if rng is not None else np.random.default_rng()

        longs, shorts = portfolio.get_longs(), portfolio.get_shorts()
        universe, longs, shorts = align_many(universe.mask, longs, shorts)
        self._universe = universe.to_numpy()
        self._long_limits = np.nansum(longs.to_numpy(), axis=1)
        self._short_limits = np.nansum(shorts.to_numpy(), axis=1)

        self._i = 0

    def __iter__(self) -> RandomPicking:
        return self

    def __next__(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        if self._i < self.n:
            self._i += 1
            return self._generate_random_picks()

        raise StopIteration

    def __len__(self) -> int:
        return self.n

    def _generate_random_picks(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        longs = np.zeros_like(self._universe, dtype=bool)
        shorts = np.zeros_like(self._universe, dtype=bool)

        for i in range(0, len(self._universe), self.holding):
            available = np.where(self._universe[i])[0]
            n_longs, n_shorts = self._long_limits[i], self._short_limits[i]

            choice = self.rng.choice(
                available,
                size=n_longs + n_shorts,
                replace=False
            )

            hold = min(len(self._universe), i + self.holding)
            longs[i:hold, choice[:n_longs]] = True
            shorts[i:hold, choice[n_longs:]] = True

        longs = pd.DataFrame(
            longs,
            index=self.portfolio.picks.index,
            columns=self.portfolio.picks.columns
        )
        shorts = pd.DataFrame(
            shorts,
            index=self.portfolio.picks.index,
            columns=self.portfolio.picks.columns
        )

        return longs, shorts


class OpportunitiesTest:
    def __init__(
            self,
            universe: Universe,
            allocation_strategy: AllocationStep | Sequence[AllocationStep],
            target_metric: Callable[[Portfolio], pd.Series],
            holding: Optional[int] = None
    ):
        self.universe = universe
        self.allocation_strategy = allocation_strategy
        self.target_metric = target_metric
        self.holding = holding

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        holding = estimate_holding(portfolio.picks) if self.holding is None else self.holding

        best_portfolio = pqr.Portfolio(
            self.universe,
            *ProphetPickingStrategy(self.universe, holding, "best")(portfolio),
            allocation_strategy=self.allocation_strategy
        )
        worst_portfolio = pqr.Portfolio(
            self.universe,
            *ProphetPickingStrategy(self.universe, holding, "worst")(portfolio),
            allocation_strategy=self.allocation_strategy
        )

        best_metric = self.target_metric(best_portfolio)
        worst_metric = self.target_metric(worst_portfolio)
        portfolio_metric = self.target_metric(portfolio)

        return (portfolio_metric - worst_metric) / (best_metric - worst_metric)


class ProphetPickingStrategy:
    def __init__(
            self,
            universe: Universe,
            holding: int,
            way: Literal["best", "worst"]
    ):
        self.universe = universe
        self.holding = holding
        self.way = way

    def __call__(self, portfolio: Portfolio) -> tuple[pd.DataFrame, pd.DataFrame]:
        longs, shorts = portfolio.get_longs(), portfolio.get_shorts()
        longs, shorts, universe, prices = align_many(longs, shorts, self.universe.mask, self.universe.prices)

        long_limits = longs.sum(axis=1)
        short_limits = shorts.sum(axis=1)

        universe_returns = prices.pct_change(self.holding).shift(-self.holding)
        universe_returns = Hold(self.holding)(universe_returns)
        universe_returns[~universe] = np.nan

        longs = pd.DataFrame(
            np.zeros_like(longs, dtype=bool),
            index=longs.index,
            columns=longs.columns
        )
        shorts = pd.DataFrame(
            np.zeros_like(shorts, dtype=bool),
            index=shorts.index,
            columns=shorts.columns
        )

        for date in universe.index:
            if self.way == "best":
                best = universe_returns.loc[date].nlargest(long_limits.loc[date]).index
                worst = universe_returns.loc[date].nsmallest(short_limits.loc[date]).index
            else:
                best = universe_returns.loc[date].nsmallest(long_limits.loc[date]).index
                worst = universe_returns.loc[date].nlargest(short_limits.loc[date]).index

            longs.loc[date, best] = True
            shorts.loc[date, worst] = True

        return longs, shorts



from __future__ import annotations

from typing import Callable, Sequence

import numpy as np
import pandas as pd

from .factor import Factor, Preprocessor
from .portfolio import Portfolio, AllocationStep
from .universe import Universe

__all__ = [
    "FactorModel",

    "Quantiles",
    "Top",
    "Bottom",
    "TimeSeries",

    "GridSearch",
]


class FactorModel:
    def __init__(
            self,
            picking_strategies: Sequence[Callable[[Factor], pd.DataFrame]],
            allocation_strategy: AllocationStep | Sequence[AllocationStep],
            add_wml: bool = False
    ):
        self.picking_strategies = picking_strategies
        self.allocation_strategy = allocation_strategy
        self.add_wml = add_wml

    def __call__(
            self,
            factor: Factor,
            universe: Universe
    ) -> list[Portfolio]:
        portfolios = [
            Portfolio(
                universe,
                longs=picker(factor),
                allocation_strategy=self.allocation_strategy,
                name=name
            )
            for picker, name in zip(
                self.picking_strategies,
                [
                    "Winners",
                    *[f"Neutral {i}" for i in range(1, len(self.picking_strategies) - 1)],
                    "Losers"
                ]
            )
        ]

        if self.add_wml:
            portfolios.append(
                Portfolio(
                    universe,
                    longs=portfolios[0].picks.astype(bool),
                    shorts=portfolios[-1].picks.astype(bool),
                    allocation_strategy=self.allocation_strategy,
                    name="WML"
                )
            )

        return portfolios


class Quantiles:
    def __init__(
            self,
            min_q: float = 0.0,
            max_q: float = 1.0
    ):
        self.min_q = min_q
        self.max_q = max_q

    def __call__(self, factor: Factor) -> pd.DataFrame:
        factor_values = factor.values.to_numpy()
        q = factor.quantile([self.min_q, self.max_q]).to_numpy()

        return pd.DataFrame(
            (q[:, [1]] <= factor_values) & (factor_values <= q[:, [0]])
            if factor.is_better_more() else
            (q[:, [0]] <= factor_values) & (factor_values <= q[:, [1]]),
            index=factor.values.index.copy(),
            columns=factor.values.columns.copy()
        )


class Top:
    def __init__(self, n: int = 10):
        self.n = n

    def __call__(self, factor: Factor) -> pd.DataFrame:
        factor_values = factor.values.to_numpy()
        top = factor.top(self.n).to_numpy()

        return pd.DataFrame(
            factor_values >= top if factor.is_better_more() else factor_values <= top,
            index=factor.values.index.copy(),
            columns=factor.values.columns.copy()
        )


class Bottom:
    def __init__(self, n: int = 10):
        self.n = n

    def __call__(self, factor: Factor) -> pd.DataFrame:
        factor_values = factor.values.to_numpy()
        bottom = factor.bottom(self.n).to_numpy()

        return pd.DataFrame(
            factor_values <= bottom if factor.is_better_more() else factor_values >= bottom,
            index=factor.values.index.copy(),
            columns=factor.values.columns.copy()
        )


class TimeSeries:
    def __init__(
            self,
            min_threshold: float = -np.inf,
            max_threshold: float = np.inf
    ):
        self.min_threshold = min_threshold
        self.max_threshold = max_threshold

    def __call__(self, factor: Factor) -> pd.DataFrame:
        factor_values = factor.values.to_numpy()

        return pd.DataFrame(
            (self.min_threshold <= factor_values) & (factor_values <= self.max_threshold),
            index=factor.values.index.copy(),
            columns=factor.values.columns.copy()
        )


class GridSearch:
    def __init__(
            self,
            factor_preprocessors: dict[str, Preprocessor | Sequence[Preprocessor]],
            factor_model: FactorModel
    ):
        self.factor_preprocessors = factor_preprocessors
        self.factor_model = factor_model

    def __call__(
            self,
            factor: Factor,
            universe: Universe,
            target: Callable[[Portfolio], float],
    ) -> pd.DataFrame:
        metrics = []

        for name, factor_preprocessor in self.factor_preprocessors.items():
            portfolios = self.factor_model(
                Factor(factor.values, factor.better, factor_preprocessor),
                universe
            )

            metrics.append(
                pd.Series(
                    {
                        portfolio.name: target(portfolio)
                        for portfolio in portfolios
                    },
                    name=name
                )
            )

        return pd.DataFrame(metrics)


def split_quantiles(n: int) -> list[Quantiles]:
    q = np.linspace(0, 1, n + 1)
    quantiles = [
        Quantiles(q[i], q[i + 1])
        for i in range(len(q) - 1)
    ]

    return quantiles



from __future__ import annotations

from dataclasses import make_dataclass
from warnings import warn

import numpy as np
import pandas as pd
import statsmodels.api as sm
from statsmodels.regression.rolling import RollingOLS

from pqr.utils import align


def extract_annualizer(df_or_series: pd.DataFrame | pd.Series) -> float:
    freq_alias = {
        "A": 1, "AS": 1, "BYS": 1, "BA": 1, "BAS": 1, "RE": 1,  # yearly
        "Q": 4, "QS": 4, "BQ": 4, "BQS": 4,  # quarterly
        "M": 12, "MS": 12, "BM": 12, "BMS": 12, "CBM": 12, "CBMS": 12,  # monthly
        "W": 52,  # weekly
        "B": 252, "C": 252, "D": 252,  # daily
    }

    if not isinstance(df_or_series.index, pd.DatetimeIndex):
        raise TypeError("df or series must have pd.DateTimeIndex to infer periodicity")

    idx = df_or_series.index
    inferred_freq = getattr(idx, "inferred_freq", None)
    annualizer = freq_alias.get(inferred_freq)

    if annualizer is None:
        warn("periodicity of df or series cannot be determined correctly, estimation is used")
        years_approx = (idx[-1] - idx[0]).days / 365.25
        annualizer = len(idx) / years_approx

    return annualizer


def adjust(
        returns: pd.Series,
        rf: float | pd.Series
) -> pd.Series:
    if isinstance(rf, pd.Series):
        returns, rf = align(returns, rf)

    return returns - rf


def stats_container_factory(metric_name: str) -> type:
    return make_dataclass(
        metric_name,
        [
            ("value", float),
            ("t_stat", float),
            ("p_value", float),
        ],
        namespace={
            "template": property(lambda self: "{value:.2f}{stars} ({t_stat:.2f})"),
            "count_stars": lambda self: 3 if self.p_value < 0.01 else (
                2 if self.p_value < 0.05 else (
                    1 if self.p_value < 0.1 else 0
                )
            )
        }
    )


def estimate_ols(
        returns: pd.Series,
        benchmark: pd.Series,
        rf: float = 0.0
):
    adjusted_returns = adjust(returns, rf)
    adjusted_benchmark = adjust(benchmark, rf)

    y, x = align(adjusted_returns, adjusted_benchmark)
    x = sm.add_constant(x.to_numpy())
    ols = sm.OLS(y.to_numpy(), x)

    return ols.fit()


def estimate_rolling_ols(
        returns: pd.Series,
        benchmark: pd.Series,
        window: int,
        rf: float = 0.0,
):
    adjusted_returns = adjust(returns, rf)
    adjusted_benchmark = adjust(benchmark, rf)

    y, x = align(adjusted_returns, adjusted_benchmark)
    x = sm.add_constant(x.to_numpy())
    ols = RollingOLS(y.to_numpy(), x, window=window)

    return ols.fit()


def estimate_holding(picks: pd.DataFrame) -> int:
    diff = np.diff(picks.to_numpy(), axis=0)
    rebalancings_long = (diff == 1).any(axis=1).sum()
    rebalancings_short = (diff == -1).any(axis=1).sum()
    avg_rebalacings = (rebalancings_long + rebalancings_short) / 2

    return round(len(diff) / avg_rebalacings)



from __future__ import annotations

from typing import Any

import numpy as np
import pandas as pd


def replace_with_nan(
        *df_or_series: pd.DataFrame | pd.Series,
        to_replace: Any = 0
) -> list[pd.DataFrame | pd.Series]:
    return [
        data.replace(to_replace, np.nan) for data in df_or_series
    ]


def is_aligned(
        left: pd.DataFrame | pd.Series,
        right: pd.DataFrame | pd.Series
) -> bool:
    if isinstance(left, pd.Series) or isinstance(right, pd.Series):
        return (
                (left.index.shape == right.index.shape) and (left.index == right.index).all()
        )

    return (
            (left.index.shape == right.index.shape) and (left.index == right.index).all()
            and
            (left.columns.shape == right.columns.shape) and (left.columns == right.columns).all()
    )


def align(
        left: pd.DataFrame | pd.Series,
        right: pd.DataFrame | pd.Series
) -> tuple[pd.DataFrame | pd.Series, pd.DataFrame | pd.Series]:
    if is_aligned(left, right):
        return left, right

    axis = None
    if isinstance(left, pd.Series) or isinstance(right, pd.Series):
        axis = 0
    return left.align(right, join="inner", axis=axis)


def align_many(*df_or_series: pd.DataFrame | pd.Series) -> tuple[pd.DataFrame | pd.Series, ...]:
    df_or_series = list(df_or_series)

    for i in range(len(df_or_series) - 1):
        df_or_series[i], df_or_series[i + 1] = align(df_or_series[i], df_or_series[i + 1])

    for i in range(len(df_or_series) - 2, 0, -1):
        df_or_series[i], df_or_series[i - 1] = align(df_or_series[i], df_or_series[i - 1])

    return tuple(df_or_series)



from __future__ import annotations

from typing import Optional, Callable

import pandas as pd

from .portfolio import Portfolio, AllocationStep
from .universe import Universe

__all__ = [
    "Benchmark",
]


class Benchmark:
    def __init__(
            self,
            returns: pd.Series,
            name: Optional[str] = None
    ):
        self.returns = returns.astype(float)
        self.name = name if name is not None else "Benchmark"
        self.returns.index.name = name

    @classmethod
    def from_index(
            cls,
            index: pd.Series,
            name: Optional[str] = None,
    ) -> Benchmark:
        return cls(
            index.pct_change().fillna(0),
            name
        )

    @classmethod
    def from_universe(
            cls,
            universe: Universe,
            weighting_strategy: Optional[AllocationStep] = None,
            name: Optional[str] = None,
    ) -> Benchmark:
        benchmark = Portfolio(
            universe,
            longs=universe.mask,
            allocation_strategy=weighting_strategy
        )

        return cls(
            benchmark.returns,
            name
        )



from __future__ import annotations

from typing import Optional, Callable, Sequence

import numpy as np
import pandas as pd

from pqr.utils import align
from .factor import Factor
from .universe import Universe
from .utils import compose, normalize

__all__ = [
    "Portfolio",
    "AllocationStep",

    "EqualWeights",
    "WeightsByFactor",
    "ScalingByFactor",
    "LeverageLimits",
    "TheoreticalAllocation",
    "CashAllocation",
]

AllocationStep = Callable[[pd.DataFrame], pd.DataFrame]


class Portfolio:
    def __init__(
            self,
            universe: Universe,
            longs: Optional[pd.DataFrame] = None,
            shorts: Optional[pd.DataFrame] = None,
            allocation_strategy: Optional[AllocationStep | Sequence[AllocationStep]] = None,
            name: Optional[str] = None
    ):
        if longs is None and shorts is None:
            raise ValueError("either longs or shorts must be specified")

        elif longs is not None and shorts is not None:  # long-short
            longs, shorts = align(longs, shorts)
            picks = pd.DataFrame(
                longs.to_numpy(dtype=np.int8) - shorts.to_numpy(dtype=np.int8),
                index=longs.index.copy(),
                columns=longs.columns.copy()
            )
        elif longs is not None:  # long-only
            picks = pd.DataFrame(
                longs.to_numpy(dtype=np.int8),
                index=longs.index.copy(),
                columns=longs.columns.copy()
            )
        else:  # short-only
            picks = pd.DataFrame(
                -shorts.to_numpy(dtype=np.int8),
                index=shorts.index.copy(),
                columns=shorts.columns.copy()
            )
        self.picks = picks

        if allocation_strategy is None:
            allocation_strategy = EqualWeights()
        elif isinstance(allocation_strategy, Sequence):
            allocation_strategy = compose(*allocation_strategy)
        self.positions = allocation_strategy(self.picks)

        self.returns = universe(self.positions)

        if name is None:
            name = "Portfolio"
        self.name = name

        self.picks.index.name = name
        self.positions.index.name = name
        self.returns.index.name = name

    def get_longs(self) -> pd.DataFrame:
        return pd.DataFrame(
            self.picks.to_numpy() == 1,
            index=self.picks.index.copy(),
            columns=self.picks.columns.copy()
        )

    def get_shorts(self) -> pd.DataFrame:
        return pd.DataFrame(
            self.picks.to_numpy() == -1,
            index=self.picks.index.copy(),
            columns=self.picks.columns.copy()
        )


class EqualWeights:
    def __call__(self, picks: pd.DataFrame) -> pd.DataFrame:
        picks_array = picks.to_numpy()
        longs, shorts = picks_array == 1, picks_array == -1

        if not longs.any() and not shorts.any():
            raise ValueError("cannot weigh portfolio without picks")

        with np.errstate(divide="ignore", invalid="ignore"):
            return pd.DataFrame(
                normalize(longs) - normalize(shorts),
                index=picks.index.copy(),
                columns=picks.columns.copy()
            )


class WeightsByFactor:
    def __init__(self, factor: Factor):
        self.factor = factor

    def __call__(self, picks: pd.DataFrame) -> pd.DataFrame:
        picks, factor_values = align(picks, self.factor.values)
        picks_array, factor_array = picks.to_numpy(), factor_values.to_numpy()
        longs, shorts = picks_array == 1, picks_array == -1

        if not longs.any() and not shorts.any():
            raise ValueError("cannot weigh portfolio without picks")

        with np.errstate(divide="ignore", invalid="ignore"):
            return pd.DataFrame(
                normalize(longs * factor_array) - normalize(shorts * factor_array),
                index=picks.index.copy(),
                columns=picks.columns.copy()
            )


class ScalingByFactor:
    def __init__(
            self,
            factor: Factor,
            target: float = 1.0
    ):
        self.factor = factor
        self.target = target

    def __call__(self, weights: pd.DataFrame) -> pd.DataFrame:
        weights, factor_values = align(weights, self.factor.values)

        if self.factor.is_better_more():
            leverage = factor_values.to_numpy() / self.target
        else:
            leverage = self.target / factor_values.to_numpy()

        return pd.DataFrame(
            leverage * weights.to_numpy(),
            index=weights.index.copy(),
            columns=weights.columns.copy()
        )


class LeverageLimits:
    def __init__(
            self,
            min_leverage: float = -np.inf,
            max_leverage: float = np.inf
    ):
        self.min_leverage = min_leverage
        self.max_leverage = max_leverage

    def __call__(self, weights: pd.DataFrame) -> pd.DataFrame:
        w = weights.to_numpy()
        total_leverage = np.nansum(w, axis=1, keepdims=True)

        with np.errstate(divide="ignore", invalid="ignore"):
            exceed_min = total_leverage < self.min_leverage
            under_min = (
                                np.where(exceed_min, w, 0) /
                                np.where(exceed_min, total_leverage, 1)
                        ) * self.min_leverage

            exceed_max = total_leverage > self.max_leverage
            above_max = (
                                np.where(exceed_max, w, 0) /
                                np.where(exceed_max, total_leverage, 1)
                        ) * self.max_leverage

        return pd.DataFrame(
            np.where(~(exceed_min | exceed_max), w, 0) +
            under_min + above_max,
            index=weights.index.copy(),
            columns=weights.columns.copy()
        )


class TheoreticalAllocation:
    def __init__(self, fee: float = 0.0):
        self.fee = fee

    def __call__(self, weights: pd.DataFrame) -> pd.DataFrame:
        w = weights.to_numpy()
        positions = pd.DataFrame(
            w * (1 - self.fee),
            index=weights.index.copy(),
            columns=weights.columns.copy()
        )
        positions.insert(
            loc=0,
            column="CASH_RESIDUALS",
            value=(1 - self.fee) - np.nansum(positions, axis=1)
        )

        return positions


class CashAllocation:
    def __init__(
            self,
            prices: pd.DataFrame,
            capital: float = 1_000_000.0,
            fee: float = 0.0
    ):
        self.prices = prices
        self.capital = capital
        self.fee = fee

    def __call__(self, weights: pd.DataFrame) -> pd.DataFrame:
        prices, weights = align(self.prices, weights)
        prices_arr = prices.to_numpy()
        weights_arr = weights.to_numpy()

        allocation = np.zeros_like(weights_arr, dtype=float)
        cash = np.ones(allocation.shape[0]) * self.capital
        balance = cash.copy()

        for i in range(len(allocation)):
            w, p, c = weights_arr[i], prices_arr[i], cash[i]
            prev_alloc = allocation[max(0, i - 1)]

            balance[i] = c + np.nansum(prev_alloc * p)

            alloc = np.nan_to_num(w * balance[i] / p).astype(int)
            alloc_diff = alloc - prev_alloc
            cash_diff = -(alloc_diff * p)
            commission = np.nansum(np.abs(cash_diff)) * self.fee

            cash[i] = c + np.nansum(cash_diff) - commission
            allocation[i] = alloc

        positions_in_cash = np.insert(
            allocation * prices_arr,
            obj=0, axis=1, values=cash,
        )

        return pd.DataFrame(
            positions_in_cash / balance[:, np.newaxis],
            index=weights.index.copy(),
            columns=["CASH_RESIDUALS"] + list(weights.columns.copy())
        )



from setuptools import setup, find_packages

with open("README.md") as f:
    readme = f.read()


setup(
    name="pqr",
    version="0.3.3",
    description="Library for testing factor strategies",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[
        "pandas",
        "numpy",
        "statsmodels",
        "scipy",
        "matplotlib",
        "IPython",
    ],
    author="eura17, atomtosov",
    license="MIT",
    license_file="LICENSE",
    platforms="any",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Office/Business :: Financial",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    test_suite="pytest",
    python_requires=">=3.7",
    project_urls={
            "Bug Tracker": "https://github.com/atomtosov/pqr/issues",
            "Documentation": "https://pqr.readthedocs.io/en/latest/index.html",
            "Source Code": "https://github.com/atomtosov/pqr/",
        },
)



from __future__ import annotations

import functools as ft
from typing import TypeVar, Callable

import numpy as np

T = TypeVar("T")


def compose(*functions: Callable[[T], T]) -> Callable[[T], T]:
    return ft.reduce(
        lambda f, g: lambda x: g(f(x)),
        functions
    )


def normalize(raw_weights: np.ndarray) -> np.ndarray:
    with np.errstate(divide="ignore", invalid="ignore"):
        return np.nan_to_num(
            raw_weights / np.nansum(raw_weights, axis=1, keepdims=True, dtype=float),
            nan=0, neginf=0, posinf=0
        )


def top_single(arr: np.ndarray, n: np.ndarray) -> np.ndarray:
    arr = np.unique(arr[~np.isnan(arr)])
    if arr.any():
        arr = np.sort(arr)
        max_len = len(arr)
        return np.array(
            [arr[-min(_n, max_len)] for _n in n]
        )
    return np.array([np.nan] * len(n))


def bottom_single(arr: np.ndarray, n: np.ndarray) -> np.ndarray:
    arr = np.unique(arr[~np.isnan(arr)])
    if arr.any():
        arr = np.sort(arr)
        max_len = len(arr)
        return np.array(
            [arr[min(_n, max_len) - 1] for _n in n]
        )
    return np.array([np.nan] * len(n))



from typing import Sequence

import numpy as np
import pandas as pd
import statsmodels.api as sm

from pqr.core.benchmark import Benchmark
from pqr.core.portfolio import Portfolio
from pqr.utils import align_many
from .utils import adjust

__all__ = [
    "FamaFrenchRegression",
    "FamaMacbethRegression",
]


class FamaFrenchRegression:
    def __init__(
            self,
            market: Benchmark,
            benchmarks: Sequence[Benchmark],
            rf: float = 0.0
    ):
        self.market = market
        self.benchmarks = benchmarks
        self.rf = rf

    def __call__(self, portfolio: Portfolio) -> pd.DataFrame:
        adjusted_returns = adjust(portfolio.returns, self.rf)
        adjusted_market_returns = adjust(self.market.returns, self.rf)

        y, *x = align_many(
            adjusted_returns,
            adjusted_market_returns,
            *(benchmark.returns for benchmark in self.benchmarks)
        )
        y = y.to_numpy()
        x = sm.add_constant(np.array(x).T)

        ols = sm.OLS(y, x)
        est = ols.fit()

        return pd.DataFrame(
            [
                [
                    est.params[i],
                    est.bse[i],
                    est.tvalues[i],
                    est.pvalues[i],
                ]
                for i in range(len(est.params))
            ],
            index=[
                "alpha",
                f"beta_{self.market.name}",
                *(f"beta_{benchmark.name}" for benchmark in self.benchmarks)
            ],
            columns=[
                "coef",
                "se",
                "t_stat",
                "p_value",
            ]
        )


class FamaMacbethRegression:
    def __init__(
            self,
            market: Benchmark,
            benchmarks: Sequence[Benchmark],
            rf: float = 0.0
    ):
        self.market = market
        self.benchmarks = benchmarks
        self.rf = rf

    def __call__(self, portfolios: Sequence[Portfolio]) -> pd.DataFrame:
        adjusted_portfolio_returns = [
            adjust(portfolio.returns, self.rf)
            for portfolio in portfolios
        ]
        adjusted_market_returns = adjust(self.market.returns, self.rf)

        betas = []
        for adjusted_returns in adjusted_portfolio_returns:
            y, *x = align_many(
                adjusted_returns,
                adjusted_market_returns,
                *(benchmark.returns for benchmark in self.benchmarks)
            )
            y = y.to_numpy()
            x = sm.add_constant(np.array(x).T)

            ols = sm.OLS(y, x)
            est = ols.fit()

            betas.append(est.params[1:])
        print(betas)
        betas_with_const = sm.add_constant(np.array(betas))
        print(betas_with_const)
        portfolio_returns = np.array(adjusted_portfolio_returns).T

        lambdas = []
        for y in portfolio_returns:
            ols = sm.OLS(y, betas_with_const)
            est = ols.fit()

            lambdas.append(est.params[1:])

        lambdas = np.array(lambdas)

        return lambdas



from __future__ import annotations

from typing import Optional, Protocol

import numpy as np
import pandas as pd
from scipy.stats import ttest_1samp

from pqr.core import Portfolio, Benchmark
from pqr.utils import align
from .utils import extract_annualizer, adjust, estimate_ols, stats_container_factory, estimate_rolling_ols

__all__ = [
    "CompoundedReturns", "Drawdown", "Turnover",
    "TotalReturn", "TrailingTotalReturn",
    "CAGR", "TrailingCAGR",
    "MeanReturn", "TrailingMeanReturn",
    "Volatility", "TrailingVolatility",
    "WinRate", "TrailingWinRate",
    "MaxDrawdown", "TrailingMaxDrawdown",
    "ValueAtRisk", "TrailingValueAtRisk",
    "ExpectedTailLoss", "TrailingExpectedTailLoss",
    "ExpectedTailReward", "TrailingExpectedTailReward",
    "RachevRatio", "TrailingRachevRatio",
    "CalmarRatio", "TrailingCalmarRatio",
    "SharpeRatio", "TrailingSharpeRatio",
    "OmegaRatio", "TrailingOmegaRatio",
    "SortinoRatio", "TrailingSortinoRatio",
    "BenchmarkCorrelation", "TrailingBenchmarkCorrelation",
    "MeanExcessReturn", "TrailingMeanExcessReturn",
    "Alpha", "TrailingAlpha",
    "Beta", "TrailingBeta",
    "MeanTurnover", "TrailingMeanTurnover",
]


class Stats(Protocol):
    value: float
    t_stat: float
    p_value: float

    def count_stars(self) -> int:
        pass

    @property
    def template(self) -> str:
        pass


class CompoundedReturns:
    def __call__(self, portfolio: Portfolio) -> pd.Series:
        return (1 + portfolio.returns).cumprod() - 1

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Compounded Returns, %"


class Drawdown:
    def __call__(self, portfolio: Portfolio) -> pd.Series:
        equity = CompoundedReturns()(portfolio) + 1
        high_water_mark = equity.cummax()
        return equity / high_water_mark - 1

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Drawdown, %"


class Turnover:
    def __call__(self, portfolio: Portfolio) -> pd.Series:
        positions = portfolio.positions.to_numpy()

        turnover = np.nansum(
            np.abs(np.diff(positions, axis=0)),
            axis=1
        )
        # add 1st period deals
        turnover = np.insert(turnover, 0, values=np.nansum(np.abs(positions[0])))

        return pd.Series(
            turnover,
            index=portfolio.positions.index
        )

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Turnover, %"


class TotalReturn:
    def __call__(self, portfolio: Portfolio) -> pd.Series:
        return CompoundedReturns()(portfolio).iat[-1]

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio) * 100, ".2f")

    @property
    def fancy_name(self) -> str:
        return "Total Return, %"


class TrailingTotalReturn:
    def __init__(self, window: Optional[int] = None):
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        return portfolio.returns.rolling(window).apply(
            lambda r: ((1 + r).cumprod() - 1).iloc[-1]
        ).iloc[window:]

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Trailing Total Return, %"


class CAGR:
    def __init__(self, annualizer: Optional[float] = None):
        self.annualizer = annualizer

    def __call__(self, portfolio: Portfolio) -> float:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        tr = TotalReturn()(portfolio)
        years = len(portfolio.returns) / annualizer

        return (1 + tr) ** (1 / years) - 1

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio) * 100, ".2f")

    @property
    def fancy_name(self) -> str:
        return "CAGR, %"


class TrailingCAGR:
    def __init__(
            self,
            annualizer: Optional[float] = None,
            window: Optional[int] = None
    ):
        self.annualizer = annualizer
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        tr = TrailingTotalReturn(window)(portfolio)
        years = window / annualizer

        return (1 + tr) ** (1 / years) - 1

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Trailing CAGR, %"


class MeanReturn:
    def __init__(
            self,
            statistics: bool = False,
            annualizer: Optional[float] = None,
    ):
        self.statistics = statistics
        self.annualizer = annualizer

    def __call__(self, portfolio: Portfolio) -> float | Stats:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        mr = portfolio.returns.mean() * annualizer

        if self.statistics:
            ttest = ttest_1samp(portfolio.returns, 0, alternative="greater")
            mr = stats_container_factory("MeanReturn")(
                value=mr,
                t_stat=ttest.statistic,
                p_value=ttest.pvalue
            )

        return mr

    def fancy(self, portfolio: Portfolio) -> str:
        mr = self(portfolio)
        if self.statistics:
            return mr.template.format(
                value=mr.value * 100,
                stars="*" * mr.count_stars(),
                t_stat=mr.t_stat
            )

        return format(mr * 100, ".2f")

    @property
    def fancy_name(self) -> str:
        return "Mean Return, %"


class TrailingMeanReturn:
    def __init__(
            self,
            statistics: bool = False,
            annualizer: Optional[float] = None,
            window: Optional[int] = None
    ):
        self.statistics = statistics
        self.annualizer = annualizer
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        return portfolio.returns.rolling(window).mean().iloc[window:] * annualizer

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Trailing Mean Return, %"


class Volatility:
    def __init__(self, annualizer: Optional[float] = None):
        self.annualizer = annualizer

    def __call__(self, portfolio: Portfolio) -> float:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        return portfolio.returns.std() * np.sqrt(annualizer)

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio) * 100, ".2f")

    @property
    def fancy_name(self) -> str:
        return "Volatility, %"


class TrailingVolatility:
    def __init__(
            self,
            annualizer: Optional[float] = None,
            window: Optional[int] = None
    ):
        self.annualizer = annualizer
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        return portfolio.returns.rolling(window).std().iloc[window:] * np.sqrt(annualizer)

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Trailing Volatility, %"


class WinRate:
    def __call__(self, portfolio: Portfolio) -> float:
        return (portfolio.returns > 0).sum() / len(portfolio.returns)

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio) * 100, ".2f")

    @property
    def fancy_name(self) -> str:
        return "Win Rate, %"


class TrailingWinRate:
    def __init__(self, window: Optional[int] = None):
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        return (portfolio.returns > 0).rolling(window).sum().iloc[window:] / window

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Trailing Win Rate, %"


class MaxDrawdown:
    def __call__(self, portfolio: Portfolio) -> float:
        return Drawdown()(portfolio).min()

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio) * 100, ".2f")

    @property
    def fancy_name(self) -> str:
        return "Maximum Drawdown, %"


class TrailingMaxDrawdown:
    def __init__(self, window: Optional[int] = None):
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        return portfolio.returns.rolling(window).apply(
            lambda r: ((1 + r).cumprod() / (1 + r).cumprod().cummax() - 1).min()
        ).iloc[window:]

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Trailing Maximum Drawdown, %"


class ValueAtRisk:
    def __init__(
            self,
            cutoff: float = 0.05,
            annualizer: Optional[float] = None
    ):
        self.cutoff = cutoff
        self.annualizer = annualizer

    def __call__(self, portfolio: Portfolio) -> float:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        return portfolio.returns.quantile(self.cutoff) * np.sqrt(annualizer)

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio) * 100, ".2f")

    @property
    def fancy_name(self) -> str:
        return "Value at Risk, %"


class TrailingValueAtRisk:
    def __init__(
            self,
            cutoff: float = 0.05,
            annualizer: Optional[float] = None,
            window: Optional[int] = None
    ):
        self.cutoff = cutoff
        self.annualizer = annualizer
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        return portfolio.returns.rolling(window).quantile(self.cutoff) * np.sqrt(annualizer)

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Trailing Value at Risk, %"


class ExpectedTailLoss:
    def __init__(
            self,
            cutoff: float = 0.05,
            annualizer: Optional[float] = None
    ):
        self.cutoff = cutoff
        self.annualizer = annualizer

    def __call__(self, portfolio: Portfolio) -> float:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        less_cutoff = portfolio.returns <= portfolio.returns.quantile(self.cutoff)
        return portfolio.returns[less_cutoff].mean() * np.sqrt(annualizer)

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio) * 100, ".2f")

    @property
    def fancy_name(self) -> str:
        return "Expected Tail Loss, %"


class TrailingExpectedTailLoss:
    def __init__(
            self,
            cutoff: float = 0.05,
            annualizer: Optional[float] = None,
            window: Optional[int] = None
    ):
        self.cutoff = cutoff
        self.annualizer = annualizer
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        return portfolio.returns.rolling(window).apply(
            lambda r: r[r <= r.quantile(self.cutoff)].mean()
        ).iloc[window:] * np.sqrt(annualizer)

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Trailing Expected Tail Loss, %"


class ExpectedTailReward:
    def __init__(
            self,
            cutoff: float = 0.95,
            annualizer: Optional[float] = None
    ):
        self.cutoff = cutoff
        self.annualizer = annualizer

    def __call__(self, portfolio: Portfolio) -> float:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        more_cutoff = portfolio.returns >= portfolio.returns.quantile(self.cutoff)
        return portfolio.returns[more_cutoff].mean() * np.sqrt(annualizer)

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio) * 100, ".2f")

    @property
    def fancy_name(self) -> str:
        return "Expected Tail Reward, %"


class TrailingExpectedTailReward:
    def __init__(
            self,
            cutoff: float = 0.95,
            annualizer: Optional[float] = None,
            window: Optional[int] = None
    ):
        self.cutoff = cutoff
        self.annualizer = annualizer
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        return portfolio.returns.rolling(window).apply(
            lambda r: r[r >= r.quantile(self.cutoff)].mean()
        ).iloc[window:] * np.sqrt(annualizer)

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Trailing Expected Tail Reward, %"


class RachevRatio:
    def __init__(
            self,
            reward_cutoff: float = 0.95,
            risk_cutoff: float = 0.05
    ):
        self.reward_cutoff = reward_cutoff
        self.risk_cutoff = risk_cutoff

    def __call__(self, portfolio: Portfolio) -> float:
        etr = ExpectedTailReward(self.reward_cutoff)(portfolio)
        etl = ExpectedTailLoss(self.risk_cutoff)(portfolio)
        return -(etr / etl)

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio), ".2f")

    @property
    def fancy_name(self) -> str:
        return "Rachev Ratio"


class TrailingRachevRatio:
    def __init__(
            self,
            reward_cutoff: float = 0.95,
            risk_cutoff: float = 0.05,
            window: Optional[int] = None
    ):
        self.reward_cutoff = reward_cutoff
        self.risk_cutoff = risk_cutoff
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        etr = TrailingExpectedTailReward(self.reward_cutoff, window)(portfolio)
        etl = TrailingExpectedTailLoss(self.risk_cutoff, window)(portfolio)
        return -(etr / etl)

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio)

    @property
    def fancy_name(self) -> str:
        return "Trailing Rachev Ratio"


class CalmarRatio:
    def __init__(self, annualizer: Optional[float] = None):
        self.annualizer = annualizer

    def __call__(self, portfolio: Portfolio) -> float:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        return -(CAGR(annualizer)(portfolio) / MaxDrawdown()(portfolio))

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio), ".2f")

    @property
    def fancy_name(self) -> str:
        return "Calmar Ratio"


class TrailingCalmarRatio:
    def __init__(
            self,
            annualizer: Optional[float] = None,
            window: Optional[int] = None
    ):
        self.annualizer = annualizer
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        return -(TrailingCAGR(annualizer, window)(portfolio) / TrailingMaxDrawdown(window)(portfolio))

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio)

    @property
    def fancy_name(self) -> str:
        return "Trailing Calmar Ratio"


class SharpeRatio:
    def __init__(
            self,
            rf: float = 0.0,
            annualizer: Optional[float] = None
    ):
        self.rf = rf
        self.annualizer = annualizer

    def __call__(self, portfolio: Portfolio) -> float:
        adjusted = adjust(portfolio.returns, self.rf)
        if self.annualizer is None:
            annualizer = extract_annualizer(adjusted)
        else:
            annualizer = self.annualizer

        return adjusted.mean() / adjusted.std() * np.sqrt(annualizer)

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio), ".2f")

    @property
    def fancy_name(self) -> str:
        return "Sharpe Ratio"


class TrailingSharpeRatio:
    def __init__(
            self,
            rf: float = 0.0,
            annualizer: Optional[float] = None,
            window: Optional[int] = None
    ):
        self.rf = rf
        self.annualizer = annualizer
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        adjusted = adjust(portfolio.returns, self.rf)
        if self.annualizer is None:
            annualizer = extract_annualizer(adjusted)
        else:
            annualizer = self.annualizer

        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        mr = adjusted.rolling(window).mean().iloc[window:]
        std = adjusted.rolling(window).std().iloc[window:]

        return mr / std * np.sqrt(annualizer)

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio)

    @property
    def fancy_name(self) -> str:
        return "Trailing Sharpe Ratio"


class OmegaRatio:
    def __init__(self, rf: float = 0.0):
        self.rf = rf

    def __call__(self, portfolio: Portfolio) -> float:
        adjusted = adjust(portfolio.returns, self.rf)
        above = adjusted[adjusted > 0].sum()
        under = adjusted[adjusted < 0].sum()
        return -(above / under)

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio), ".2f")

    @property
    def fancy_name(self) -> str:
        return "Omega Ratio"


class TrailingOmegaRatio:
    def __init__(
            self,
            rf: float = 0.0,
            window: Optional[int] = None
    ):
        self.rf = rf
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        adjusted = adjust(portfolio.returns, self.rf)

        return adjusted.rolling(window).apply(
            lambda r: -(r[r > 0].sum() / r[r < 0].sum())
        )

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio)

    @property
    def fancy_name(self) -> str:
        return "Omega Ratio"


class SortinoRatio:
    def __init__(
            self,
            rf: float = 0.0,
            annualizer: Optional[float] = None
    ):
        self.rf = rf
        self.annualizer = annualizer

    def __call__(self, portfolio: Portfolio) -> float:
        adjusted = adjust(portfolio.returns, self.rf)
        if self.annualizer is None:
            annualizer = extract_annualizer(adjusted)
        else:
            annualizer = self.annualizer

        returns_under_mar = np.clip(adjusted, a_min=-np.inf, a_max=0)
        downside_risk = np.sqrt((returns_under_mar ** 2).mean())

        return adjusted.mean() / downside_risk * np.sqrt(annualizer)

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio), ".2f")

    @property
    def fancy_name(self) -> str:
        return "Sortino Ratio"


class TrailingSortinoRatio:
    def __init__(
            self,
            rf: float = 0.0,
            annualizer: Optional[float] = None,
            window: Optional[int] = None
    ):
        self.rf = rf
        self.annualizer = annualizer
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        adjusted = adjust(portfolio.returns, self.rf)
        if self.annualizer is None:
            annualizer = extract_annualizer(adjusted)
        else:
            annualizer = self.annualizer

        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        mr = adjusted.rolling(window).mean().iloc[window:]
        downside_risk = adjusted.rolling(window).apply(
            lambda r: np.sqrt((np.clip(adjusted, a_min=-np.inf, a_max=0) ** 2).mean())
        ).iloc[window:]

        return mr / downside_risk * np.sqrt(annualizer)

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio)

    @property
    def fancy_name(self) -> str:
        return "Trailing Sortino Ratio"


class BenchmarkCorrelation:
    def __init__(self, benchmark: Benchmark):
        self.benchmark = benchmark

    def __call__(self, portfolio: Portfolio) -> float:
        returns, benchmark = align(portfolio.returns, self.benchmark.returns)
        return returns.corr(benchmark)

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio), ".2f")

    @property
    def fancy_name(self) -> str:
        return f"{self.benchmark.name} Correlation"


class TrailingBenchmarkCorrelation:
    def __init__(
            self,
            benchmark: Benchmark,
            window: Optional[int] = None
    ):
        self.benchmark = benchmark
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        returns, benchmark = align(portfolio.returns, self.benchmark.returns)

        if self.window is None:
            window = int(extract_annualizer(returns))
        else:
            window = self.window

        return returns.rolling(window).corr(benchmark)

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio)

    @property
    def fancy_name(self) -> str:
        return f"Trailing {self.benchmark.name} Correlation"


class MeanExcessReturn:
    def __init__(
            self,
            benchmark: Benchmark,
            statistics: bool = False,
            annualizer: Optional[float] = None
    ):
        self.benchmark = benchmark
        self.statistics = statistics
        self.annualizer = annualizer

    def __call__(self, portfolio: Portfolio) -> float | Stats:
        adjusted = adjust(portfolio.returns, self.benchmark.returns)
        if self.annualizer is None:
            annualizer = extract_annualizer(adjusted)
        else:
            annualizer = self.annualizer

        mer = adjusted.mean() * annualizer

        if self.statistics:
            ttest = ttest_1samp(portfolio.returns, 0, alternative="greater")
            mer = stats_container_factory("MeanExcessReturn")(
                value=mer,
                t_stat=ttest.statistic,
                p_value=ttest.pvalue
            )

        return mer

    def fancy(self, portfolio: Portfolio) -> str:
        mer = self(portfolio)

        if self.statistics:
            return mer.template.format(
                value=mer.value * 100,
                stars="*" * mer.count_stars(),
                t_stat=mer.t_stat
            )

        return format(mer * 100, ".2f")

    @property
    def fancy_name(self) -> str:
        return "Mean Excess Return, %"


class TrailingMeanExcessReturn:
    def __init__(
            self,
            benchmark: Benchmark,
            annualizer: Optional[float] = None,
            window: Optional[int] = None
    ):
        self.benchmark = benchmark
        self.annualizer = annualizer
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        adjusted = adjust(portfolio.returns, self.benchmark.returns)

        if self.annualizer is None:
            annualizer = extract_annualizer(adjusted)
        else:
            annualizer = self.annualizer

        if self.window is None:
            window = int(extract_annualizer(adjusted))
        else:
            window = self.window

        return adjusted.rolling(window).mean().iloc[window:] * annualizer

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Trailing Mean Excess Return, %"


class Alpha:
    def __init__(
            self,
            benchmark: Benchmark,
            rf: float = 0.0,
            statistics: bool = False,
            annualizer: Optional[float] = None
    ):
        self.benchmark = benchmark
        self.rf = rf
        self.statistics = statistics
        self.annualizer = annualizer

    def __call__(self, portfolio: Portfolio) -> float | Stats:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        returns, benchmark = align(portfolio.returns, self.benchmark.returns)
        est = estimate_ols(returns, benchmark, self.rf)
        alpha = est.params[0] * annualizer

        if self.statistics:
            # TODO: t-stat and p-value for one-sided test
            alpha = stats_container_factory("Alpha")(
                value=alpha,
                p_value=est.pvalues[0],
                t_stat=est.tvalues[0]
            )

        return alpha

    def fancy(self, portfolio: Portfolio) -> str:
        alpha = self(portfolio)

        if self.statistics:
            return alpha.template.format(
                value=alpha.value * 100,
                stars="*" * alpha.count_stars(),
                t_stat=alpha.t_stat
            )

        return format(alpha * 100, ".2f")

    @property
    def fancy_name(self) -> str:
        return "Alpha, %"


class TrailingAlpha:
    def __init__(
            self,
            benchmark: Benchmark,
            rf: float = 0.0,
            annualizer: Optional[float] = None,
            window: Optional[int] = None
    ):
        self.benchmark = benchmark
        self.rf = rf
        self.annualizer = annualizer
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        returns, benchmark = align(portfolio.returns, self.benchmark.returns)

        return pd.Series(
            estimate_rolling_ols(
                returns,
                benchmark,
                window,
                self.rf
            ).params[window:, 0] * annualizer,
            index=returns.index[window:].copy()
        )

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Trailing Alpha, %"


class Beta:
    def __init__(
            self,
            benchmark: Benchmark,
            rf: float = 0.0,
            statistics: bool = False,
    ):
        self.benchmark = benchmark
        self.rf = rf
        self.statistics = statistics

    def __call__(self, portfolio: Portfolio) -> float | Stats:
        returns, benchmark = align(portfolio.returns, self.benchmark.returns)
        est = estimate_ols(returns, benchmark, self.rf)
        beta = est.params[1]

        if self.statistics:
            beta = stats_container_factory("Beta")(
                value=beta,
                p_value=est.pvalues[1],
                t_stat=est.tvalues[1]
            )

        return beta

    def fancy(self, portfolio: Portfolio) -> str:
        beta = self(portfolio)

        if self.statistics:
            return beta.template.format(
                value=beta.value,
                stars="*" * beta.count_stars(),
                t_stat=beta.t_stat
            )

        return format(beta, ".2f")

    @property
    def fancy_name(self) -> str:
        return "Beta"


class TrailingBeta:
    def __init__(
            self,
            benchmark: Benchmark,
            rf: float = 0.0,
            window: Optional[int] = None
    ):
        self.benchmark = benchmark
        self.rf = rf
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        returns, benchmark = align(portfolio.returns, self.benchmark.returns)

        return pd.Series(
            estimate_rolling_ols(
                returns, benchmark, window, self.rf
            ).params[window:, 1],
            index=returns.index[window:].copy()
        )

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio)

    @property
    def fancy_name(self) -> str:
        return "Trailing Beta, %"


class MeanTurnover:
    def __init__(self, annualizer: Optional[float] = None):
        self.annualizer = annualizer

    def __call__(self, portfolio: Portfolio) -> float:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        return Turnover()(portfolio).mean() * annualizer

    def fancy(self, portfolio: Portfolio) -> str:
        return format(self(portfolio) * 100, ".2f")

    @property
    def fancy_name(self) -> str:
        return "Mean Turnover, %"


class TrailingMeanTurnover:
    def __init__(
            self,
            annualizer: Optional[float] = None,
            window: Optional[int] = None
    ):
        self.annualizer = annualizer
        self.window = window

    def __call__(self, portfolio: Portfolio) -> pd.Series:
        if self.annualizer is None:
            annualizer = extract_annualizer(portfolio.returns)
        else:
            annualizer = self.annualizer

        if self.window is None:
            window = int(extract_annualizer(portfolio.returns))
        else:
            window = self.window

        return Turnover()(portfolio).rolling(window).mean().iloc[window:] * annualizer

    def fancy(self, portfolio: Portfolio) -> pd.Series:
        return self(portfolio) * 100

    @property
    def fancy_name(self) -> str:
        return "Trailing Mean Turnover, %"
